package net.wanji.opt.service.impl;

import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.ObjectUtil;
import lombok.extern.slf4j.Slf4j;
import net.wanji.common.enums.CrossStatusEnum;
import net.wanji.common.enums.TurnConvertEnum;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.common.utils.tool.TimeArrayUtil;
import net.wanji.databus.bo.CrossIdAndStartEndDateBO;
import net.wanji.databus.bo.CrossIdBO;
import net.wanji.databus.dao.entity.BaseCrossDirInfoPO;
import net.wanji.databus.dao.entity.BaseCrossSchemePO;
import net.wanji.databus.dao.entity.CrossSectionPO;
import net.wanji.databus.dao.mapper.*;
import net.wanji.databus.po.*;
import net.wanji.databus.vo.RunningEvaluateCrossListVO;
import net.wanji.opt.bo.CrossNameBO;
import net.wanji.opt.bo.MetricsDetailBO;
import net.wanji.opt.service.RunningEvaluateService;
import net.wanji.opt.vo.*;
import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * @author Kent HAN
 * @date 2023/8/21 9:53
 */
@Slf4j
@Service
public class RunningEvaluateServiceImpl implements RunningEvaluateService {
    private final CrossDataRealtimeMapper crossDataRealtimeMapper;
    private final CrossDataHistMapper crossDataHistMapper;
    private final BaseCrossSectionMapper baseCrossSectionMapper;
    private final BaseCrossSchemeMapper baseCrossSchemeMapper;
    private final BaseCrossDirInfoMapper baseCrossDirInfoMapper;
    private final BaseCrossTurnInfoMapper baseCrossTurnInfoMapper;
    private final LaneInfoMapper laneInfoMapper;
    private final BaseCrossInfoMapper baseCrossInfoMapper;

    public RunningEvaluateServiceImpl(CrossDataRealtimeMapper crossDataRealtimeMapper,
                                      CrossDataHistMapper crossDataHistMapper,
                                      @Qualifier("baseCrossSectionMapper") BaseCrossSectionMapper baseCrossSectionMapper,
                                      @Qualifier("baseCrossSchemeMapper") BaseCrossSchemeMapper baseCrossSchemeMapper,
                                      @Qualifier("baseCrossDirInfoMapper") BaseCrossDirInfoMapper baseCrossDirInfoMapper,
                                      @Qualifier("baseCrossTurnInfoMapper") BaseCrossTurnInfoMapper baseCrossTurnInfoMapper,
                                      @Qualifier("laneInfoMapper") LaneInfoMapper laneInfoMapper,
                                      @Qualifier("baseCrossInfoMapper") BaseCrossInfoMapper baseCrossInfoMapper) {
        this.crossDataRealtimeMapper = crossDataRealtimeMapper;
        this.crossDataHistMapper = crossDataHistMapper;
        this.baseCrossSectionMapper = baseCrossSectionMapper;
        this.baseCrossSchemeMapper = baseCrossSchemeMapper;
        this.baseCrossDirInfoMapper = baseCrossDirInfoMapper;
        this.baseCrossTurnInfoMapper = baseCrossTurnInfoMapper;
        this.laneInfoMapper = laneInfoMapper;
        this.baseCrossInfoMapper = baseCrossInfoMapper;
    }

    @Override
    public List<RunningEvaluateCrossListVO> crossList(CrossNameBO crossNameBO) {
        String crossName = crossNameBO.getCrossName();
        List<RunningEvaluateCrossListVO> crossDataRealtimePOList = crossDataRealtimeMapper.selectByCrossName(crossName);
        for (RunningEvaluateCrossListVO runningEvaluateCrossListVO : crossDataRealtimePOList) {
            Double sturation = runningEvaluateCrossListVO.getSturation();
            runningEvaluateCrossListVO.setServiceLevel(CrossUtil.getServiceLevel(sturation));
            // 路口经纬度
            String crossId = runningEvaluateCrossListVO.getCrossId();
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(crossId);
            String location = baseCrossInfoPO.getLocation();
            double[] lonLat = CrossUtil.getLonLat(location);
            runningEvaluateCrossListVO.setLongitude(lonLat[0]);
            runningEvaluateCrossListVO.setLatitude(lonLat[1]);
        }
        // 按服务水平降序排序
        List<RunningEvaluateCrossListVO> res = crossDataRealtimePOList.stream()
                .sorted(Comparator.comparing(RunningEvaluateCrossListVO::getSturation).reversed())
                .collect(Collectors.toList());

        return res;
    }

    @Override
    public RunningEvaluateCrossEvaluateVO crossEvaluate(CrossIdAndStartEndDateBO bo) {
        RunningEvaluateCrossEvaluateVO vo = new RunningEvaluateCrossEvaluateVO();

        String crossId = bo.getCrossId();
        List<CrossDataHistPO> crossDataHistPOList = buildCrossDataHistPOList(bo, crossId);

        CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
        Double sturation = crossDataRealtimePO.getSturation();
        String serviceLevel = CrossUtil.getServiceLevel(sturation);
        vo.setServiceLevel(serviceLevel);


        List<CrossDataHistPO> congestionEventList = buildCongestionEvents(crossDataHistPOList); // 获取拥堵事件集合
        vo.setCongestionTimes(congestionEventList.size());
        List<CrossDataHistPO> unbalanceEventList = buildUnbalanceEvents(crossDataHistPOList); // 获取失衡事件集合
        vo.setUnbalanceTimes(unbalanceEventList.size());
        List<CrossDataHistPO> spilloverEventList = buildSpilloverEvents(crossDataHistPOList);  // 获取溢出事件集合
        vo.setSpilloverTimes(spilloverEventList.size());
        Integer schemeProblems = calcSchemeProblems(
                congestionEventList, unbalanceEventList, spilloverEventList, crossId); // 计算方案问题
        vo.setSchemeProblems(schemeProblems);

        return vo;
    }

    private List<CrossDataHistPO> buildCrossDataHistPOList(CrossIdAndStartEndDateBO bo, String crossId) {
        Date startDate = bo.getStartDate();
        Date endDate = bo.getEndDate();
        endDate = DateUtil.offsetDay(endDate, 1); // 包含最后一天
        int startStamp = (int) (startDate.getTime() / 1000); // 10位时间戳
        int endStamp = (int) (endDate.getTime() / 1000);
        List<CrossDataHistPO> crossDataHistPOList = crossDataHistMapper.selectByCrossIdAndStartEnd(
                crossId, startStamp, endStamp);
        return crossDataHistPOList;
    }

    @Override
    public List<RunningEvaluateIndexStatusVO> congestionStatus(CrossIdAndStartEndDateBO bo) {

        String crossId = bo.getCrossId();
        int status = CrossStatusEnum.CONGESTION.getCode();

        return buildRes(bo, crossId, status);
    }

    @Override
    public List<RunningEvaluateIndexStatusVO> unbalanceStatus(CrossIdAndStartEndDateBO bo) {
        String crossId = bo.getCrossId();
        int status = CrossStatusEnum.UNBALANCE.getCode();

        return buildRes(bo, crossId, status);
    }

    @Override
    public List<RunningEvaluateIndexStatusVO> spilloverStatus(CrossIdAndStartEndDateBO bo) {
        String crossId = bo.getCrossId();
        int status = CrossStatusEnum.SPILLOVER.getCode();

        return buildRes(bo, crossId, status);
    }

    @Override
    public List<RunningEvaluateSchemeProblemsVO> schemeProblems(CrossIdAndStartEndDateBO bo) {
        String crossId = bo.getCrossId();
        List<RunningEvaluateSchemeProblemsVO> res = new ArrayList<>();
        List<String> timeArray = TimeArrayUtil.getTimeArray();

        for (String time : timeArray) {
            RunningEvaluateSchemeProblemsVO vo = new RunningEvaluateSchemeProblemsVO();
            vo.setTime(time);
            List<CrossDataHistPO> crossDataHistPOList = buildCrossDataHistPOList(bo, crossId);
            // 找出有问题的记录
            List<CrossDataHistPO> collect = crossDataHistPOList.stream()
                    .filter(po -> !Objects.equals(po.getStatus(), CrossStatusEnum.NORMAL.getCode()))
                    .collect(Collectors.toList());
            // 按采集时间小时进行聚合
            SimpleDateFormat sdf = new SimpleDateFormat("HH");
            Map<String, List<CrossDataHistPO>> timePOMap = crossDataHistPOList.stream()
                    .collect(Collectors.groupingBy(po -> {
                        Date date = new Date((long) po.getBatchTime() * 1000); // 将10位时间戳转换为Date对象
                        return sdf.format(date); // 将Date对象转换为"HH"格式的字符串
                    }));
            // 找出当前小时的记录
            List<CrossDataHistPO> currentHourList = timePOMap.get(time.substring(0, 2)); // 保留小时字符串
            if (currentHourList != null) {
                // 按相同开始时间去重，保留持续时间最长的一个
                List<CrossDataHistPO> uniqueList = currentHourList.stream()
                        .collect(Collectors.groupingBy(
                                CrossDataHistPO::getStartTime, // 按startTime分组
                                Collectors.collectingAndThen(
                                        // 在每个组中找到duration最大的元素
                                        Collectors.maxBy(Comparator.comparingLong(CrossDataHistPO::getDuration)),
                                        Optional::get // 从Optional中获取元素
                                )
                        ))
                        .values().stream().collect(Collectors.toList()); // 从Map中获取值并收集到新的列表中
                // 统计数量
                vo.setCounts(uniqueList.size());
                // 找出出现问题的日期和方案
                List<RunningEvaluateSchemeProblemsVO.DateAndScheme> dateAndSchemeList = new ArrayList<>();
                dateAndSchemeList = buildDateAndSchemeList(uniqueList, crossId);
                vo.setDateAndSchemeList(dateAndSchemeList);
            }

            res.add(vo);
        }
        return res;
    }

    @Override
    public List<RunningEvaluateMetricsDetailVO> metricsDetail(MetricsDetailBO bo) {
        return null;
    }

    @Override
    public RunningEvaluateScopeTreeVO scopeTree(CrossIdBO bo) {
        String crossId = bo.getCrossId();
        RunningEvaluateScopeTreeVO runningEvaluateScopeTreeVO = new RunningEvaluateScopeTreeVO();
        runningEvaluateScopeTreeVO.setCrossId(crossId);
        List<RunningEvaluateScopeTreeVO.DirVO> dirVOList= buildDirVOList(crossId);
        runningEvaluateScopeTreeVO.setChild(dirVOList);
        return runningEvaluateScopeTreeVO;
    }

    private List<RunningEvaluateScopeTreeVO.DirVO> buildDirVOList(String crossId) {
        List<RunningEvaluateScopeTreeVO.DirVO> res = new ArrayList<>();
        Integer type = 1; // 进口
        List<BaseCrossDirInfoPO> baseCrossDirInfoPOList =
                baseCrossDirInfoMapper.selectByCrossIdAndInOutType(crossId, type);

        for (BaseCrossDirInfoPO baseCrossDirInfoPO : baseCrossDirInfoPOList) {
            RunningEvaluateScopeTreeVO.DirVO dirVO = new RunningEvaluateScopeTreeVO.DirVO();
            Integer dir = baseCrossDirInfoPO.getDirType();
            dirVO.setDir(dir);
            List<RunningEvaluateScopeTreeVO.TurnVO> turnVOList = buildTurnVOList(crossId, dir);
            dirVO.setChild(turnVOList);
            res.add(dirVO);
        }

        return res;
    }

    private List<RunningEvaluateScopeTreeVO.TurnVO> buildTurnVOList(String crossId, Integer dir) {
        List<RunningEvaluateScopeTreeVO.TurnVO> res = new ArrayList<>();
        List<CrossTurnInfoPO> crossTurnInfoPOList = baseCrossTurnInfoMapper.selectByCrossIdAndDir(crossId, dir);

        for (CrossTurnInfoPO crossTurnInfoPO : crossTurnInfoPOList) {
            RunningEvaluateScopeTreeVO.TurnVO turnVO = new RunningEvaluateScopeTreeVO.TurnVO();
            String turnType = crossTurnInfoPO.getTurnType();
            turnVO.setTurn(turnType);
            List<RunningEvaluateScopeTreeVO.LaneVO> laneVOList = buildLaneVOList(crossId, dir, turnType);
            turnVO.setChild(laneVOList);
            res.add(turnVO);
        }

        return res;
    }

    private List<RunningEvaluateScopeTreeVO.LaneVO> buildLaneVOList(String crossId, Integer dir, String turnType) {
        List<RunningEvaluateScopeTreeVO.LaneVO> res = new ArrayList<>();
        List<Integer> keyList = TurnConvertEnum.getKeyListBySingleCode(turnType);
        Integer type = 2; // 进口
        if (ObjectUtil.isNotEmpty(keyList)) {
            List<LaneInfoPO> laneInfoPOList = laneInfoMapper.selectByTurnType(crossId, type, dir, keyList);
            for (LaneInfoPO laneInfoPO : laneInfoPOList) {
                RunningEvaluateScopeTreeVO.LaneVO laneVO = new RunningEvaluateScopeTreeVO.LaneVO();
                Integer sort = laneInfoPO.getSort();
                laneVO.setLaneSort(sort);
                res.add(laneVO);
            }
        }
        return res;
    }

    private List<RunningEvaluateSchemeProblemsVO.DateAndScheme> buildDateAndSchemeList(
            List<CrossDataHistPO> uniqueList, String crossId) {
        List<RunningEvaluateSchemeProblemsVO.DateAndScheme> res = new ArrayList<>();
        // 将问题按日期分组
        Map<LocalDate, List<CrossDataHistPO>> groupedByDate = uniqueList.stream()
                .collect(Collectors.groupingBy(
                        // 将10位时间戳转换为Instant
                        crossDataHistPO -> Instant.ofEpochSecond(crossDataHistPO.getBatchTime())
                                .atZone(ZoneId.systemDefault()) // 将Instant转换为系统默认时区的ZonedDateTime
                                .toLocalDate() // 获取LocalDate
                ));

        for (Map.Entry<LocalDate, List<CrossDataHistPO>> entry : groupedByDate.entrySet()) {
            LocalDate localDate = entry.getKey(); // 获取日期
            List<CrossDataHistPO> crossDataHistPOs = entry.getValue(); // 获取该日期下的CrossDataHistPO对象列表

            RunningEvaluateSchemeProblemsVO.DateAndScheme dateAndScheme =
                    new RunningEvaluateSchemeProblemsVO.DateAndScheme();
            Date date = Date.from(localDate.atStartOfDay()
                    .atZone(ZoneId.systemDefault())
                    .toInstant());
            dateAndScheme.setProblemDate(date);
            Set<String> schemeNameSet = new HashSet<>();

            for (CrossDataHistPO abnormalEvent : crossDataHistPOs) {
                SimpleDateFormat sdf = new SimpleDateFormat("HH:mm");
                Date abnormalEventStartTime = abnormalEvent.getStartTime();
                Integer duration = abnormalEvent.getDuration();
                Date abnormalEventEndTime = DateUtil.offsetMinute(abnormalEventStartTime, duration);
                String eventStartTime = sdf.format(abnormalEventStartTime);
                String eventEndTime = sdf.format(abnormalEventEndTime);

                List<CrossSectionPO> crossSectionPOList = baseCrossSectionMapper.selectByCrossId(crossId);
                for (CrossSectionPO crossSectionPO : crossSectionPOList) {
                    String sectionStartTime = crossSectionPO.getStartTime();
                    String sectionEndTime = crossSectionPO.getEndTime();
                    // 比较两组时间段是否相交
                    boolean isIntersecting = isTimeIntersecting(
                            eventStartTime, eventEndTime, sectionStartTime, sectionEndTime);
                    if (isIntersecting) {
                        Integer schemeId = crossSectionPO.getSchemeId();
                        BaseCrossSchemePO baseCrossSchemePO = baseCrossSchemeMapper.selectById(schemeId);
                        schemeNameSet.add(baseCrossSchemePO.getName());
                        dateAndScheme.setSchemeNameSet(schemeNameSet);

                    }
                }
            }
            res.add(dateAndScheme);
        }
        return res;
    }

    @NotNull
    private List<RunningEvaluateIndexStatusVO> buildRes(CrossIdAndStartEndDateBO bo, String crossId, int status) {
        List<CrossDataHistPO> crossDataHistPOList = buildCrossDataHistPOList(bo, crossId);
        List<CrossDataHistPO> collect = crossDataHistPOList.stream()
                .filter(po -> po.getStatus() == status)
                .collect(Collectors.toList());

        List<RunningEvaluateIndexStatusVO> res = buildRunningEvaluateIndexStatusVOList(collect, status);
        return res;
    }

    @NotNull
    private List<RunningEvaluateIndexStatusVO> buildRunningEvaluateIndexStatusVOList(
            List<CrossDataHistPO> crossDataHistPOList, int status) {
        SimpleDateFormat sdf = new SimpleDateFormat("HH");
        Map<String, List<CrossDataHistPO>> timePOMap = crossDataHistPOList.stream()
                .collect(Collectors.groupingBy(po -> {
                    Date date = new Date((long) po.getBatchTime() * 1000); // 将10位时间戳转换为Date对象
                    return sdf.format(date); // 将Date对象转换为"HH"格式的字符串
                }));

        // 构造结果集
        List<RunningEvaluateIndexStatusVO> res = new ArrayList<>();
        List<String> timeArray = TimeArrayUtil.getTimeArray();
        for (String time : timeArray) {
            RunningEvaluateIndexStatusVO vo = new RunningEvaluateIndexStatusVO();
            vo.setTime(time);
            List<CrossDataHistPO> subList = timePOMap.get(time.substring(0, 2)); // 保留小时字符串
            if (subList != null) {
                Double avgCongestionIndex = calcAvgIndex(subList, status);
                vo.setIndex(avgCongestionIndex);
            }

            res.add(vo);
        }
        return res;
    }

    private Double calcAvgIndex(List<CrossDataHistPO> subList, int status) {
        Double average = 0.0;

        if (status == CrossStatusEnum.CONGESTION.getCode()) {
            average = subList.stream()
                    .mapToDouble(CrossDataHistPO::getCongestionIndex)
                    .average()
                    .orElse(0.0);
        } else if (status == CrossStatusEnum.UNBALANCE.getCode()) {
            average = subList.stream()
                    .mapToDouble(CrossDataHistPO::getUnbalanceIndex)
                    .average()
                    .orElse(0.0);
        } else if (status == CrossStatusEnum.SPILLOVER.getCode()) {
            average = subList.stream()
                    .mapToDouble(CrossDataHistPO::getQueueLength)
                    .average()
                    .orElse(0.0);
        }

        DecimalFormat df = new DecimalFormat("#.00");
        String formattedAverage = df.format(average);
        return Double.parseDouble(formattedAverage);
    }

    private Integer calcSchemeProblems(List<CrossDataHistPO> congestionEventList,
                                       List<CrossDataHistPO> unbalanceEventList,
                                       List<CrossDataHistPO> spilloverEventList,
                                       String crossId) {

        Set<Integer> schemeIdSet = new HashSet<>(); // 用于方案ID去重

        List<CrossDataHistPO> allList = Stream.of(congestionEventList, unbalanceEventList, spilloverEventList)
                .flatMap(List::stream)
                .collect(Collectors.toList());

        for (CrossDataHistPO abnormalEvent : allList) {
            SimpleDateFormat sdf = new SimpleDateFormat("HH:mm");
            Date abnormalEventStartTime = abnormalEvent.getStartTime();
            Integer duration = abnormalEvent.getDuration();
            Date abnormalEventEndTime = DateUtil.offsetMinute(abnormalEventStartTime, duration);
            String eventStartTime = sdf.format(abnormalEventStartTime);
            String eventEndTime = sdf.format(abnormalEventEndTime);

            List<CrossSectionPO> crossSectionPOList = baseCrossSectionMapper.selectByCrossId(crossId);
            for (CrossSectionPO crossSectionPO : crossSectionPOList) {
                String sectionStartTime = crossSectionPO.getStartTime();
                String sectionEndTime = crossSectionPO.getEndTime();
                // 比较两组时间段是否相交
                boolean isIntersecting = isTimeIntersecting(
                        eventStartTime, eventEndTime, sectionStartTime, sectionEndTime);
                if (isIntersecting) {
                    Integer schemeId = crossSectionPO.getSchemeId();
                    schemeIdSet.add(schemeId);
                }
            }
        }

        return schemeIdSet.size();
    }

    private boolean isTimeIntersecting(String startTime1, String endTime1, String startTime2, String endTime2) {
        return !(endTime1.compareTo(startTime2) <= 0 || startTime1.compareTo(endTime2) >= 0);
    }


    private List<CrossDataHistPO> buildSpilloverEvents(List<CrossDataHistPO> crossDataHistPOList) {
        return buildEvents(crossDataHistPOList, CrossStatusEnum.SPILLOVER.getCode());
    }

    private List<CrossDataHistPO> buildUnbalanceEvents(List<CrossDataHistPO> crossDataHistPOList) {
        return buildEvents(crossDataHistPOList, CrossStatusEnum.UNBALANCE.getCode());
    }

    private List<CrossDataHistPO> buildCongestionEvents(List<CrossDataHistPO> crossDataHistPOList) {
        return buildEvents(crossDataHistPOList, CrossStatusEnum.CONGESTION.getCode());
    }

    private List<CrossDataHistPO> buildEvents(List<CrossDataHistPO> crossDataHistPOList, int status) {

        List<CrossDataHistPO> collect = crossDataHistPOList.stream()
                .filter(po -> po.getStatus() == status)
                .collect(Collectors.groupingBy(
                        CrossDataHistPO::getStartTime,
                        Collectors.collectingAndThen(
                                Collectors.maxBy(Comparator.comparingInt(CrossDataHistPO::getDuration)),
                                Optional::get
                        )
                ))
                .values()
                .stream()
                .collect(Collectors.toList());

        return collect;
    }

}
